/*
 *  Code extracted from
 *  linux/kernel/hd.c
 *
 *  Copyright (C) 1991-1998  Linus Torvalds
 *
 *  devfs support - jj, rgooch, 980122
 *
 *  Moved partition checking code to fs/partitions* - Russell King
 *  (linux@arm.uk.linux.org)
 */

/*
 * TODO:  rip out the remaining init crap from this file  --hch
 */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/fs.h>
#include <linux/genhd.h>
#include <linux/kernel.h>
#include <linux/blk.h>
#include <linux/init.h>
#include <linux/spinlock.h>


static rwlock_t gendisk_lock;

/*
 * Global kernel list of partitioning information.
 *
 * XXX: you should _never_ access this directly.
 *	the only reason might ever be exported is source compatiblity.
 */
static struct gendisk *gendisk_head;


/**
 * add_gendisk - add partitioning information to kernel list
 * @gp: per-device partitioning information
 *
 * This function registers the partitioning information in @gp
 * with the kernel.
 */
void
add_gendisk(struct gendisk *gp)
{
	struct gendisk *sgp;
	unsigned long flags;

	write_lock_irqsave(&gendisk_lock, flags);

	/*
 	 *	In 2.5 this will go away. Fix the drivers who rely on
 	 *	old behaviour.
 	 */

	for (sgp = gendisk_head; sgp; sgp = sgp->next)
	{
		if (sgp == gp)
		{
#if 0
			printk(KERN_ERR
			       "add_gendisk: device major %d is buggy "
			       "and added a live gendisk!\n",
				sgp->major)
#endif
			goto out;
		}
	}

	major_gendisk[gp->major] = gp;
	gp->next = gendisk_head;
	gendisk_head = gp;
out:
	write_unlock_irqrestore(&gendisk_lock, flags);
}

EXPORT_SYMBOL(add_gendisk);


/**
 * del_gendisk - remove partitioning information from kernel list
 * @gp: per-device partitioning information
 *
 * This function unregisters the partitioning information in @gp
 * with the kernel.
 */
void
del_gendisk(struct gendisk *gp)
{
	struct gendisk **gpp;
	unsigned long flags;

	write_lock_irqsave(&gendisk_lock, flags);
	major_gendisk[gp->major] = NULL;
	for (gpp = &gendisk_head; *gpp; gpp = &((*gpp)->next))
		if (*gpp == gp)
			break;
	if (*gpp)
		*gpp = (*gpp)->next;
	write_unlock_irqrestore(&gendisk_lock, flags);
}

EXPORT_SYMBOL(del_gendisk);


/**
 * get_gendisk - get partitioning information for a given device
 * @dev: device to get partitioning information for
 *
 * This function gets the structure containing partitioning
 * information for the given device @dev.
 */
struct gendisk *
get_gendisk(kdev_t dev)
{
	struct gendisk *gp = NULL;
	int maj = MAJOR(dev);

	read_lock(&gendisk_lock);
	for (gp = gendisk_head; gp; gp = gp->next)
		if (gp->major == maj)
			break;
	read_unlock(&gendisk_lock);

	return gp;
}

EXPORT_SYMBOL(get_gendisk);


#define MSEC(x) (((x) / HZ) * 1000 + ((x) % HZ) * 1000 / HZ)

#ifdef CONFIG_PROC_FS
int
get_partition_list(char *page, char **start, off_t offset, int count)
{
	struct gendisk *gp;
	struct hd_struct *hd;
	char buf[64];
	int len, n;

	len = sprintf(page, "major minor  #blocks  name     "
		 "rio rmerge rsect ruse wio wmerge "
	  	"wsect wuse running use aveq\n\n");
		
		
	read_lock(&gendisk_lock);
	for (gp = gendisk_head; gp; gp = gp->next) {
		for (n = 0; n < (gp->nr_real << gp->minor_shift); n++) {
			if (gp->part[n].nr_sects == 0)
				continue;

				 hd = &gp->part[n]; disk_round_stats(hd);
                                len += sprintf(page + len,
                                              "%4d  %4d %10d %s "
                                              "%u %u %u %u %u %u %u %u %u %u %u\n",
                                              gp->major, n, gp->sizes[n],
                                              disk_name(gp, n, buf),
                                              hd->rd_ios, hd->rd_merges,
                                              hd->rd_sectors,
                                              MSEC(hd->rd_ticks),
                                              hd->wr_ios, hd->wr_merges,
                                              hd->wr_sectors,
                                              MSEC(hd->wr_ticks),
                                              hd->ios_in_flight,
                                              MSEC(hd->io_ticks),
                                              MSEC(hd->aveq));

			if (len < offset)
				offset -= len, len = 0;
			else if (len >= offset + count)
				goto out;
		}
	}

out:
	read_unlock(&gendisk_lock);
	*start = page + offset;
	len -= offset;
	if (len < 0)
		len = 0;
	return len > count ? count : len;
}
#endif


extern int blk_dev_init(void);
#ifdef CONFIG_BLK_DEV_DAC960
extern void DAC960_Initialize(void);
#endif
extern int net_dev_init(void);
extern void console_map_init(void);
extern int atmdev_init(void);
extern int i2o_init(void);
extern int cpqarray_init(void);

int __init device_init(void)
{
	rwlock_init(&gendisk_lock);
	blk_dev_init();
	sti();
#ifdef CONFIG_I2O
	i2o_init();
#endif
#ifdef CONFIG_BLK_DEV_DAC960
	DAC960_Initialize();
#endif
#ifdef CONFIG_BLK_CPQ_DA
	cpqarray_init();
#endif
#ifdef CONFIG_NET
	net_dev_init();
#endif
#ifdef CONFIG_ATM
	(void) atmdev_init();
#endif
#ifdef CONFIG_VT
	console_map_init();
#endif
	return 0;
}

__initcall(device_init);
